import sys

import numpy as np
from psygnal import Signal
from useq import MDAEvent, MDASequence

__all__ = [
    "_CMMCoreSignaler",
    "_get_auto_callback_class",
]


class _CMMCoreSignaler:
    """Signals that will be emitted from CMMCorePlus and RemoteMMCore objects."""

    # native MMCore callback events
    propertiesChanged = Signal()
    propertyChanged = Signal(str, str, str)
    channelGroupChanged = Signal(str)
    configGroupChanged = Signal(str, str)
    configSet = Signal(str, str)
    systemConfigurationLoaded = Signal()
    pixelSizeChanged = Signal(float)
    pixelSizeAffineChanged = Signal(float, float, float, float, float, float)
    stagePositionChanged = Signal(str, float)
    XYStagePositionChanged = Signal(str, float, float)
    exposureChanged = Signal(str, float)
    SLMExposureChanged = Signal(str, float)

    # added for CMMCorePlus
    sequenceStarted = Signal(MDASequence)  # at the start of an MDA sequence
    sequencePauseToggled = Signal(bool)  # when MDA is paused/unpaused
    sequenceCanceled = Signal(MDASequence)  # when mda is canceled
    sequenceFinished = Signal(MDASequence)  # when mda is done (whether canceled or not)
    frameReady = Signal(np.ndarray, MDAEvent)  # after each event in the sequence
    imageSnapped = Signal(np.ndarray)  # whenever snap is called

    # aliases for lower casing
    @property
    def xYStagePositionChanged(self):
        return self.XYStagePositionChanged

    @property
    def sLMExposureChanged(self):
        return self.SLMExposureChanged


def _get_auto_callback_class():
    for modname in {"PyQt5", "PySide2", "PyQt6", "PySide6"}:
        qmodule = sys.modules.get(modname)
        if qmodule:
            QtWidgets = getattr(qmodule, "QtWidgets")
            if QtWidgets.QApplication.instance() is not None:
                from .qcallback import QCoreCallback

                return QCoreCallback

    return _CMMCoreSignaler



import numpy as np
import pytest
from useq import MDAEvent, MDASequence

from pymmcore_plus.client import RemoteMMCore
from pymmcore_plus.client.callbacks.basic import SynchronousCallback
from pymmcore_plus.client.callbacks.qcallback import QCoreCallback
from pymmcore_plus.server import DEFAULT_URI


@pytest.fixture
def proxy():
    with RemoteMMCore() as mmcore:
        mmcore.loadSystemConfiguration()
        yield mmcore


def test_client(proxy):
    assert str(proxy._pyroUri) == DEFAULT_URI
    proxy.getConfigGroupState("Channel")


def test_mda(qtbot, proxy):
    mda = MDASequence(time_plan={"interval": 0.1, "loops": 2})

    def _check_frame(img, event):
        return (
            isinstance(img, np.ndarray)
            and isinstance(event, MDAEvent)
            and event.sequence == mda
            and event.sequence is not mda
        )

    def _check_seq(obj):
        return obj.uid == mda.uid

    signals = [
        proxy.events.sequenceStarted,
        proxy.events.frameReady,
        proxy.events.frameReady,
        proxy.events.sequenceFinished,
    ]
    signals = [
        (proxy.events.sequenceStarted, "started"),
        (proxy.events.frameReady, "frameReady1"),
        (proxy.events.frameReady, "frameReady2"),
        (proxy.events.sequenceFinished, "finishd"),
    ]
    checks = [_check_seq, _check_frame, _check_frame, _check_seq]

    with qtbot.waitSignals(signals, check_params_cbs=checks, order="strict"):
        proxy.run_mda(mda)


# test canceling while waiting for the next time point
def test_mda_cancel(qtbot, proxy: RemoteMMCore):
    mda = MDASequence(time_plan={"interval": 5000, "loops": 3})
    with qtbot.waitSignal(proxy.events.sequenceStarted):
        proxy.run_mda(mda)
    with qtbot.waitSignals(
        [proxy.events.sequenceFinished, proxy.events.sequenceCanceled], timeout=500
    ):
        proxy.cancel()


# TODO: this test may accidentally pass if qtbot is created before this
@pytest.mark.xfail(
    reason="synchronous callbacks not working yet", raises=AssertionError, strict=True
)
def test_cb_without_qt(proxy):
    """This tests that we can call a core method within a callback

    currently only works for Qt callbacks... need to figure out synchronous approach.
    """
    assert isinstance(proxy.events, SynchronousCallback)
    cam = [None]

    @proxy.events.systemConfigurationLoaded.connect
    def _cb():
        cam[0] = proxy.getCameraDevice()

    proxy.loadSystemConfiguration()
    assert cam[0] == "Camera"


def test_cb_with_qt(qtbot, proxy):
    """This tests that we can call a core method within a callback

    currently only works for Qt callbacks... need to figure out synchronous approach.
    """
    # because we're running with qt active
    assert isinstance(proxy.events, QCoreCallback)
    cam = [None]

    @proxy.events.systemConfigurationLoaded.connect
    def _cb():
        cam[0] = proxy.getCameraDevice()

    with qtbot.waitSignal(proxy.events.systemConfigurationLoaded):
        proxy.loadSystemConfiguration()
    assert cam[0] == "Camera"



"""pythonic wrapper on pymmcore.Configuration object."""
from __future__ import annotations

from collections import defaultdict
from typing import Any, DefaultDict, Dict, Iterator, Tuple

import pymmcore

_NULL = object()


# class PropertySetting(pymmcore.PropertySetting):
#     """Encompasses a device label, property name, and property value"""

#     # pymmcore API:
#     # def getDeviceLabel(self) -> str:  # i.e. 'Camera'
#     # def getKey(self) -> str:  # ie. 'Camera-Binning'
#     # def getPropertyName(self) -> str:  # ie. 'Binning'
#     # def getPropertyValue(self) -> str:  # ie. '1'
#     # def getReadOnly(self) -> bool:
#     # def getVerbose(self) -> str:  # ie. 'Camera:Binning=1'
#     # def isEqualTo(self, ps: PropertySetting) -> bool:# devLabel, propName & value eq

#     @classmethod
#     def from_property_setting(cls, ps: pymmcore.PropertySetting) -> PropertySetting:
#         label = ps.getDeviceLabel()
#         prop = ps.getPropertyName()
#         value = ps.getPropertyValue()
#         readOnly = ps.getReadOnly()
#         return cls(label, prop, value, readOnly)

#     def __repr__(self) -> str:
#         return f"<PropertySetting '{self}'>"

#     def __str__(self) -> str:
#         return self.getVerbose().replace(":=", "")

#     def __iter__(self) -> Iterable[str]:
#         yield self.getDeviceLabel()
#         yield self.getPropertyName()
#         yield self.getPropertyValue()


class Configuration(pymmcore.Configuration):
    """Encapsulation of the configuration information, with convenience methods.

    This pymmcore_plus variant provides additional conveniences:
        __len__ - number of settings
        __str__ - pretty printing of Config
        __contains__ - check if (devLabel, propLabel) is in the config
        __getitem__ - get property setting by index or (devLabel, propLabel) key
        __iter__ - iterate over (devLabeL, propLabel, value) tuples
        dict() - convert Configuration to nested dict
        json() - convert to JSON string
        yaml() - convert to YAML string (requires PyYAML)
        html() - convert to HTML string
    """

    # pymmcore API:
    # def __init__(self) -> None: ...
    # def addSetting(self, setting: PropertySetting) -> None: ...
    # def deleteSetting(self, device: str, prop, str) -> None: ...
    # @overload
    # def getSetting(self, index: int) -> PropertySetting: ...
    # @overload
    # def getSetting(self, device: str, prop: str) -> PropertySetting: ...
    # def getVerbose(self) -> str: ...
    # def isConfigurationIncluded(self, cfg: Configuration) -> bool: ...
    # def isPropertyIncluded(self, device: str, prop: str) -> bool: ...
    # def isSettingIncluded(self, ps: PropertySetting) -> bool: ...
    # def size(self) -> int: ...

    def __len__(self) -> int:
        return self.size()

    def __repr__(self) -> str:
        return f"<MMCore Configuration with {self.size()} settings>"

    def __str__(self) -> str:
        lines = []
        for device, prop in self.dict().items():
            lines.append(f"{device}:")
            for name, value in prop.items():
                lines.append(f"  {name}={value}")
            lines.append("")
        return "\n".join(lines)

    def __iter__(self) -> Iterator[Tuple[str, str, str]]:
        for i in range(self.size()):
            ps = self.getSetting(i)
            yield ps.getDeviceLabel(), ps.getPropertyName(), ps.getPropertyValue()

    def __getitem__(self, key):
        """get property setting by index or (devLabel, propLabel) key"""
        if isinstance(key, int):
            return self.getSetting(key).getPropertyValue()
        if isinstance(key, tuple):
            return self.getSetting(*key).getPropertyValue()
        raise TypeError("key must be either an int or 2-tuple of strings.")

    def __contains__(self, query):
        if isinstance(query, pymmcore.Configuration):
            return self.isConfigurationIncluded(query)
        if isinstance(query, pymmcore.PropertySetting):
            return self.isSettingIncluded(query)
        if (
            not isinstance(query, (list, tuple))
            or len(query) != 2
            or not all(isinstance(i, str) for i in query)
        ):
            raise TypeError(
                "Configuration.__contains__ expects a Configuration, a PropertySetting,"
                " or a 2-tuple of (deviceLabel, propertyLabel)"
            )
        return self.isPropertyIncluded(*query)

    def html(self) -> str:
        """Return config as HTML."""
        return self.getVerbose()

    def dict(self) -> Dict[str, Dict[str, str]]:
        """Return config as a nested dict"""
        d: DefaultDict[str, Dict[str, str]] = defaultdict(dict)
        for label, prop, value in self:
            d[label][prop] = value
        return dict(d)

    def json(self) -> str:
        """Dump config to JSON string."""
        from json import dumps

        return dumps(self.dict())

    def yaml(self) -> str:
        """Dump config to YAML string (requires PyYAML)."""
        try:
            from yaml import safe_dump
        except ImportError:  # pragma: no cover
            raise ImportError("Could not import yaml.  Please `pip install PyYAML`.")

        return safe_dump(self.dict())

    @classmethod
    def from_configuration(cls, config: pymmcore.Configuration):
        """Create Configuration (Plus) from pymmcore.Configuration"""
        new = cls()
        for s in range(config.size()):
            new.addSetting(config.getSetting(s))
        return new

    @classmethod
    def create(cls, *args, **kwargs):
        """More flexible init to create a Configuration from:

        1. A dict of dicts (outer key is device, inner key is prop)
        2. A sequence of 3-tuple
        3. kwargs: where the key is the device, and the value is a {prop: value} map
        """
        cfg = cls()
        if args:
            if len(args) > 1:  # pragma: no cover
                raise ValueError(
                    f"create takes 1 positional argument but {len(args)} were given"
                )

            arg = args[0]
            err_msg = "Argument must be either a dict of dicts, or a list of 3-tuple"
            props = []
            if isinstance(arg, dict):
                kwargs = {**arg, **kwargs}
            elif isinstance(arg, (tuple, list)):
                for item in arg:
                    if len(item) != 3:
                        raise ValueError(err_msg)
                    cfg.addSetting(pymmcore.PropertySetting(*(str(x) for x in item)))
            else:
                raise ValueError()
        if kwargs:
            for dev_label, props in kwargs.items():
                if not isinstance(props, dict):
                    raise ValueError(err_msg)
                for prop, value in props.items():
                    cfg.addSetting(
                        pymmcore.PropertySetting(dev_label, prop, str(value))
                    )
        return cfg

    def __eq__(self, o: Any) -> bool:
        if not isinstance(o, Configuration):
            return False
        return o.dict() == self.dict()



__all__ = [
    "CORE_NAME",
    "DEFAULT_HOST",
    "DEFAULT_PORT",
    "DEFAULT_URI",
    "pyroCMMCore",
    "serve",
    "try_kill_server",
]

from ._pyrocore import pyroCMMCore
from ._server import (
    CORE_NAME,
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_URI,
    serve,
    try_kill_server,
)



try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

from .client import RemoteMMCore
from .core import (
    ActionType,
    CMMCorePlus,
    Configuration,
    DeviceDetectionStatus,
    DeviceNotification,
    DeviceType,
    FocusDirection,
    Metadata,
    PortType,
    PropertyType,
)

__all__ = [
    "ActionType",
    "CMMCorePlus",
    "Configuration",
    "DeviceDetectionStatus",
    "DeviceNotification",
    "DeviceType",
    "FocusDirection",
    "Metadata",
    "PortType",
    "PropertyType",
    "RemoteMMCore",
    "CMMCorePlus",
]



import operator

import numpy as np
import pymmcore
from useq import MDAEvent

from pymmcore_plus import Configuration, Metadata
from pymmcore_plus._serialize import (
    SerCMMError,
    SerConfiguration,
    SerMDAEvent,
    SerMetadata,
    SerNDArray,
)


def _roundtrip(serializer, obj, compare=operator.eq):
    return compare(serializer().from_dict("", serializer().to_dict(obj)), obj)


def test_ndarray():
    assert _roundtrip(SerNDArray, np.random.rand(4, 4), np.allclose)


def test_cmmerror():
    assert _roundtrip(
        SerCMMError, pymmcore.CMMError("msg", 1), lambda a, b: a.getMsg() == b.getMsg()
    )


def test_mda():
    assert _roundtrip(SerMDAEvent, MDAEvent(exposure=42))


def test_configiration():
    cfg = Configuration.create({"a": {"a0": "0", "a1": "1"}, "b": {"b0": "10"}})
    assert _roundtrip(SerConfiguration, cfg)


def test_metadata():
    assert _roundtrip(SerMetadata, Metadata({"a": "1"}))



from Pyro5.api import expose
from qtpy.QtCore import QObject, Signal


class QCoreCallback(QObject):

    # native MMCore callback events
    propertiesChanged = Signal()
    propertyChanged = Signal(str, str, object)
    channelGroupChanged = Signal(str)
    configGroupChanged = Signal(str, str)
    configSet = Signal(str, str)
    systemConfigurationLoaded = Signal()
    pixelSizeChanged = Signal(float)
    pixelSizeAffineChanged = Signal(float, float, float, float, float, float)
    stagePositionChanged = Signal(str, float)
    XYStagePositionChanged = Signal(str, float, float)
    xYStagePositionChanged = XYStagePositionChanged  # alias
    exposureChanged = Signal(str, float)
    SLMExposureChanged = Signal(str, float)
    sLMExposureChanged = SLMExposureChanged  # alias

    # added for CMMCorePlus
    sequenceStarted = Signal(object)  # at the start of an MDA sequence
    sequencePauseToggled = Signal(bool)  # when MDA is paused/unpaused
    sequenceCanceled = Signal(object)  # when mda is canceled
    sequenceFinished = Signal(object)  # when mda is done (whether canceled or not)
    frameReady = Signal(object, object)  # after each event in the sequence
    imageSnapped = Signal(object)  # after an image is snapped

    @expose
    def receive_core_callback(self, signal_name, args):
        # let it throw an exception.
        getattr(self, signal_name).emit(*args)



from __future__ import annotations

import atexit
import os
import re
import time
import weakref
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from textwrap import dedent
from threading import RLock, Thread
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Pattern,
    Tuple,
    TypeVar,
    Union,
)

import pymmcore
from loguru import logger
from psygnal import SignalInstance
from wrapt import synchronized

from .._util import find_micromanager
from ._config import Configuration
from ._constants import DeviceDetectionStatus, DeviceType, PropertyType
from ._metadata import Metadata
from ._signals import _CMMCoreSignaler, _get_auto_callback_class

if TYPE_CHECKING:
    import numpy as np
    from useq import MDASequence

_T = TypeVar("_T")

ListOrTuple = Union[List[_T], Tuple[_T, ...]]

_OBJECTIVE_DEVICE_RE = re.compile(
    "(.+)?(nosepiece|obj(ective)?)(turret)?s?", re.IGNORECASE
)
_CHANNEL_REGEX = re.compile("(chan{1,2}(el)?|filt(er)?)s?", re.IGNORECASE)


@contextmanager
def _blockSignal(obj, signal):
    if isinstance(signal, SignalInstance):
        signal.block()
        yield
        signal.unblock()
    else:
        obj.blockSignals(True)
        yield
        obj.blockSignals(False)


_instance = None


class CMMCorePlus(pymmcore.CMMCore):
    lock = RLock()

    @classmethod
    def instance(
        cls, mm_path=None, adapter_paths: ListOrTuple[str] = ()
    ) -> CMMCorePlus:
        global _instance
        if _instance is None:
            _instance = cls(mm_path, adapter_paths)
        return _instance

    def __init__(self, mm_path=None, adapter_paths: ListOrTuple[str] = ()):
        super().__init__()

        self._mm_path = mm_path or find_micromanager()
        if not adapter_paths and self._mm_path:
            adapter_paths = [self._mm_path]
        if adapter_paths:
            self.setDeviceAdapterSearchPaths(adapter_paths)

        self.events = _get_auto_callback_class()()
        self._callback_relay = MMCallbackRelay(self.events)
        self.registerCallback(self._callback_relay)
        self._canceled = False
        self._paused = False

        self._objective_regex = _OBJECTIVE_DEVICE_RE
        self._channel_group_regex = _CHANNEL_REGEX

        # use weakref to avoid atexit keeping us from being
        # garbage collected
        self._weak_clean = weakref.WeakMethod(self.unloadAllDevices)
        atexit.register(self._weak_clean)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} at {hex(id(self))}>"

    def __del__(self):
        atexit.unregister(self._weak_clean)
        self.unloadAllDevices()

    # Re-implemented methods from the CMMCore API

    @synchronized(lock)
    def setProperty(
        self, label: str, propName: str, propValue: Union[bool, float, int, str]
    ) -> None:
        """setProperty with more reliable event emission.

        As stated by Nico: "Callbacks are mainly used to give devices the opportunity to
        signal back to the UI."
        https://forum.image.sc/t/micromanager-events-core-events-not-coming-through/53014/2

        Because it's left to the device adapter to emit a signal, in many cases uses
        `setProperty()` will NOT lead to a new `propertyChanged` event getting emitted.
        But that makes it hard to create listeners (i.e. in the gui or elsewhere).

        While this method override cannot completely solve that problem (core-internal
        changes will still lack an associated event emission in many cases), it can at
        least guarantee that if we use `CMMCorePlus.setProperty` to change the property,
        then a `propertyChanged` event will be emitted if the value did indeed change.

        Parameters
        ----------
        label : str
            device label
        propName : str
            property name
        propValue : Union[bool, float, int, str]
            new value
        """
        before = super().getProperty(label, propName)
        with _blockSignal(
            self.events, self.events.propertyChanged
        ):  # block the native event.
            super().setProperty(label, propName, propValue)
        after = super().getProperty(label, propName)
        if before != after:
            self.events.propertyChanged.emit(label, propName, after)

    def setDeviceAdapterSearchPaths(self, adapter_paths: ListOrTuple[str]) -> None:
        # add to PATH as well for dynamic dlls
        if (
            not isinstance(adapter_paths, (list, tuple))
            and adapter_paths
            and all(isinstance(i, str) for i in adapter_paths)
        ):
            raise TypeError("adapter paths must be a sequence of strings")
        env_path = os.environ["PATH"]
        for p in adapter_paths:
            if p not in env_path:
                env_path = p + os.pathsep + env_path
        os.environ["PATH"] = env_path
        logger.info(f"setting adapter search paths: {adapter_paths}")
        super().setDeviceAdapterSearchPaths(adapter_paths)

    @synchronized(lock)
    def loadSystemConfiguration(
        self, fileName: str | Path = "MMConfig_demo.cfg"
    ) -> None:
        """Load a config file.

        For relative paths first checks relative to the current
        working directory, then in the device adapter path.
        """
        fpath = Path(fileName).expanduser()
        if not fpath.exists() and not fpath.is_absolute() and self._mm_path:
            fpath = Path(self._mm_path) / fileName
        if not fpath.exists():
            raise FileNotFoundError(f"Path does not exist: {fpath}")
        logger.debug("loading config")
        super().loadSystemConfiguration(str(fpath.resolve()))
        logger.debug("config loaded")

    def unloadAllDevices(self) -> None:
        # this log won't appear when exiting ipython
        # but the method is still called
        logger.info("Unloading all devices")
        return super().unloadAllDevices()

    def getDeviceType(self, label: str) -> DeviceType:
        """Returns device type."""
        return DeviceType(super().getDeviceType(label))

    def getPropertyType(self, label: str, propName: str) -> PropertyType:
        return PropertyType(super().getPropertyType(label, propName))

    def detectDevice(self, deviceLabel: str) -> DeviceDetectionStatus:
        """Tries to communicate to a device through a given serial port.

        Used to automate discovery of correct serial port.
        Also configures the serial port correctly.
        """
        return DeviceDetectionStatus(super().detectDevice(deviceLabel))

    # config overrides

    def getConfigData(
        self, configGroup: str, configName: str, *, native=False
    ) -> Configuration:
        """Returns the configuration object for a given group and name."""

        cfg = super().getConfigData(configGroup, configName)
        return cfg if native else Configuration.from_configuration(cfg)

    def getPixelSizeConfigData(self, configName: str, *, native=False) -> Configuration:
        """Returns the configuration object for a given pixel size preset."""
        cfg = super().getPixelSizeConfigData(configName)
        return cfg if native else Configuration.from_configuration(cfg)

    def getConfigGroupState(self, group: str, *, native=False) -> Configuration:
        """Returns the partial state of the system, for the devices included in the
        specified group.
        """
        cfg = super().getConfigGroupState(group)
        return cfg if native else Configuration.from_configuration(cfg)

    def getConfigGroupStateFromCache(
        self, group: str, *, native=False
    ) -> Configuration:
        """Returns the partial state of the system cache, for the devices included
        in the specified group.
        """
        cfg = super().getConfigGroupStateFromCache(group)
        return cfg if native else Configuration.from_configuration(cfg)

    def getConfigState(self, group: str, config: str, *, native=False) -> Configuration:
        """Returns a partial state of the system, for devices included in the
        specified configuration.
        """
        cfg = super().getConfigState(group, config)
        return cfg if native else Configuration.from_configuration(cfg)

    def getSystemState(self, *, native=False) -> Configuration:
        """Returns the entire system state."""
        cfg = super().getSystemState()
        return cfg if native else Configuration.from_configuration(cfg)

    def getSystemStateCache(self, *, native=False) -> Configuration:
        """Returns the entire system state from cache"""
        cfg = super().getSystemStateCache()
        return cfg if native else Configuration.from_configuration(cfg)

    # metadata overloads that don't require instantiating metadata first

    @synchronized(lock)
    def getLastImageMD(
        self, md: Optional[Metadata] = None
    ) -> Tuple[np.ndarray, Metadata]:
        if md is None:
            md = Metadata()
        img = super().getLastImageMD(md)
        return img, md

    @synchronized(lock)
    def popNextImageMD(
        self, md: Optional[Metadata] = None
    ) -> Tuple[np.ndarray, Metadata]:
        if md is None:
            md = Metadata()
        img = super().popNextImageMD(md)
        return img, md

    @synchronized(lock)
    def popNextImage(self) -> np.ndarray:
        """Gets and removes the next image from the circular buffer.

        The pymmcore-plus implementation will convert images with n_components > 1
        to a shape (w, h, num_components) and dtype `img.dtype.itemsize//ncomp`
        """
        return self._fix_image(super().popNextImage())

    @synchronized(lock)
    def getNBeforeLastImageMD(
        self, n: int, md: Optional[Metadata] = None
    ) -> Tuple[np.ndarray, Metadata]:
        if md is None:
            md = Metadata()
        img = super().getNBeforeLastImageMD(n, md)
        return img, md

    def setConfig(self, groupName: str, configName: str) -> None:
        """Applies a configuration to a group."""
        super().setConfig(groupName, configName)
        # The onConfigGroupChanged callback has some limitations as
        # discussed in https://github.com/micro-manager/mmCoreAndDevices/issues/25
        # use the pymmcore-plus configSet signal as a workaround
        self.events.configSet.emit(groupName, configName)

    # NEW methods

    def getDeviceProperties(self, device_label: str) -> Dict[str, Any]:
        """Return all current properties for device `device_label`."""
        return {
            name: self.getProperty(device_label, name)
            for name in self.getDevicePropertyNames(device_label)
        }

    def getDeviceSchema(self, device_label: str) -> Dict[str, Any]:
        """Return dict in JSON-schema format for propties of `device_label`.

        Use `json.dump` to convert this dict to a JSON string.
        """
        d = {
            "title": self.getDeviceName(device_label),
            "description": self.getDeviceDescription(device_label),
            "type": "object",
            "properties": {},
        }
        for prop_name in self.getDevicePropertyNames(device_label):
            _type = self.getPropertyType(device_label, prop_name)
            d["properties"][prop_name] = p = {}
            if _type.to_json() != "null":
                p["type"] = _type.to_json()
            if self.hasPropertyLimits(device_label, prop_name):
                min_ = self.getPropertyLowerLimit(device_label, prop_name)
                max_ = self.getPropertyUpperLimit(device_label, prop_name)
                p["minimum"] = min_
                p["maximum"] = max_
            allowed = self.getAllowedPropertyValues(device_label, prop_name)
            if allowed:
                if set(allowed) == {"0", "1"} and _type.to_json() == "integer":
                    p["type"] = "boolean"
                else:
                    cls = _type.to_python()
                    p["enum"] = [cls(i) if cls else i for i in allowed]
            if self.isPropertyReadOnly(device_label, prop_name):
                p["readOnly"] = True
                p["default"] = self.getProperty(device_label, prop_name)
            if self.isPropertySequenceable(device_label, prop_name):
                p["sequenceable"] = True
                p["sequence_max_length"] = self.getPropertySequenceMaxLength(
                    device_label, prop_name
                )
            if self.isPropertyPreInit(device_label, prop_name):
                p["preInit"] = True
        if not d["properties"]:
            del d["properties"]
            del d["type"]
        return d

    @property
    def objective_device_pattern(self):
        return self._objective_regex

    @objective_device_pattern.setter
    def objective_device_pattern(self, value: Union[Pattern, str]):
        if isinstance(value, str):
            value = re.compile(value, re.IGNORECASE)
        elif not isinstance(value, Pattern):
            raise TypeError(
                "Objective Pattern must be a string or compiled regex"
                f" but is type {type(value)}"
            )
        self._objective_regex = value

    @property
    def channelGroup_pattern(self):
        return self._channelGroup_regex

    @channelGroup_pattern.setter
    def channelGroup_pattern(self, value: Union[Pattern, str]):
        if isinstance(value, str):
            value = re.compile(value, re.IGNORECASE)
        elif not isinstance(value, Pattern):
            raise TypeError(
                "channelGroup Pattern must be a string or compiled regex"
                f"but is type {type(value)}"
            )
        self._channel_group_regex = value

    def guessObjectiveDevices(self) -> List[str]:
        """
        Find any loaded devices that are likely to be an Objective/Nosepiece.

        Likely matches are loaded StateDevices with names that match this object's
        ``objective_device_pattern`` property. This is a settable property
        with a default value of::

            re.compile("(.+)?(nosepiece|obj(ective)?)(turret)?s?", re.IGNORECASE)``
        """
        devices = []

        for device in self.getLoadedDevicesOfType(DeviceType.StateDevice):
            if self._objective_regex.match(device):
                devices.append(device)
        return devices

    def getOrGuessChannelGroup(self) -> List[str]:
        """
        Get the channelGroup or find a likely set of candidates.

        If the group is not defined via ``.getChannelGroup`` then likely candidates
        will be found by searching for config groups with names that match this
        object's ``channelGroup_pattern`` property. This is a settable property
        with a default value of::

            reg = re.compile("(chan{1,2}(el)?|filt(er)?)s?", re.IGNORECASE)

        """
        chan_group = self.getChannelGroup()
        if chan_group:
            return [chan_group]
        # not set in core. Try "Channel" and other variations as fallbacks
        channel_guess = []
        for group in self.getAvailableConfigGroups():
            if self._channel_group_regex.match(group):
                channel_guess.append(group)
        return channel_guess

    def setRelativeXYZPosition(
        self, dx: float = 0, dy: float = 0, dz: float = 0
    ) -> None:
        """Sets the relative XYZ position in microns."""
        if dx or dy:
            x, y = self.getXPosition(), self.getYPosition()
            self.setXYPosition(x + dx, y + dy)
        if dz:
            z = self.getPosition(self.getFocusDevice())
            self.setZPosition(z + dz)
        self.waitForDevice(self.getXYStageDevice())
        self.waitForDevice(self.getFocusDevice())

    def getZPosition(self) -> float:
        return self.getPosition(self.getFocusDevice())

    def setZPosition(self, val: float) -> None:
        return self.setPosition(self.getFocusDevice(), val)

    @synchronized(lock)
    def setPosition(self, deviceLabel: str, val: float) -> None:
        return super().setPosition(deviceLabel, val)

    @synchronized(lock)
    def setXYPosition(self, x: float, y: float) -> None:
        return super().setXYPosition(x, y)

    @synchronized(lock)
    def getCameraChannelNames(self) -> Tuple[str, ...]:
        return tuple(
            self.getCameraChannelName(i)
            for i in range(self.getNumberOfCameraChannels())
        )

    @synchronized(lock)
    def snapImage(self) -> None:
        return super().snapImage()

    def run_mda(self, sequence: MDASequence) -> Thread:
        th = Thread(target=self._mda, args=(sequence,))
        th.start()
        return th

    def _mda(self, sequence) -> None:
        self.events.sequenceStarted.emit(sequence)
        logger.info("MDA Started: {}", sequence)
        self._paused = False
        paused_time = 0.0
        t0 = time.perf_counter()  # reference time, in seconds

        def check_canceled():
            if self._canceled:
                logger.warning("MDA Canceled: {}", sequence)
                self.events.sequenceCanceled.emit(sequence)
                self._canceled = False
                return True
            return False

        for event in sequence:
            while self._paused and not self._canceled:
                paused_time += 0.1  # fixme: be more precise
                time.sleep(0.1)

            if check_canceled():
                break

            if event.min_start_time:
                go_at = event.min_start_time + paused_time
                # We need to enter a loop here checking paused and canceled.
                # otherwise you'll potentially wait a long time to cancel
                to_go = go_at - (time.perf_counter() - t0)
                while to_go > 0:
                    while self._paused and not self._canceled:
                        paused_time += 0.1  # fixme: be more precise
                        to_go += 0.1
                        time.sleep(0.1)

                    if self._canceled:
                        break
                    if to_go > 0.5:
                        time.sleep(0.5)
                    else:
                        time.sleep(to_go)
                    to_go = go_at - (time.perf_counter() - t0)

            # check canceled again in case it was canceled
            # during the waiting loop
            if check_canceled():
                break

            logger.info(event)

            # prep hardware
            if event.x_pos is not None or event.y_pos is not None:
                x = event.x_pos or self.getXPosition()
                y = event.y_pos or self.getYPosition()
                self.setXYPosition(x, y)
            if event.z_pos is not None:
                self.setZPosition(event.z_pos)
            if event.channel is not None:
                self.setConfig(event.channel.group, event.channel.config)
            if event.exposure is not None:
                self.setExposure(event.exposure)

            # acquire
            self.waitForSystem()
            self.snapImage()
            img = self.getImage()

            self.events.frameReady.emit(img, event)

        logger.info("MDA Finished: {}", sequence)
        self.events.sequenceFinished.emit(sequence)

    def _fix_image(self, img: np.ndarray) -> np.ndarray:
        """Fix img shape/dtype based on `self.getNumberOfComponents()`.

        convert images with n_components > 1
        to a shape (w, h, num_components) and dtype `img.dtype.itemsize//ncomp`

        Parameters
        ----------
        img : np.ndarray
            input image

        Returns
        -------
        np.ndarray
            output image (possibly new shape and dtype)
        """
        if self.getNumberOfComponents() == 4:
            new_shape = img.shape + (4,)
            img = img.view(dtype=f"u{img.dtype.itemsize//4}")
            img = img.reshape(new_shape)[:, :, (2, 1, 0, 3)]  # mmcore gives bgra
        return img

    def snap(self, *args, fix=True) -> np.ndarray:
        """
        snap and return an image.

        In contrast to ``snapImage`` this will directly return the image
        without also calling ``getImage``.

        Parameters
        ----------
        *args :
            Passed through to ``getImage``
        fix : bool, default: True
            Whether to fix the shape of images with n_components >1
            Pass on to ``getImage``

        Returns
        -------
        img : np.ndarray
        """
        self.snapImage()
        img = self.getImage()
        self.events.imageSnapped.emit(img)
        return img

    def getImage(self, *args, fix=True) -> np.ndarray:
        """Exposes the internal image buffer.

        The pymmcore-plus implementation will convert images with n_components > 1
        to a shape (w, h, num_components) and dtype `img.dtype.itemsize//ncomp`
        """
        img = super().getImage(*args)
        return self._fix_image(img) if fix else img

    def cancel(self):
        self._canceled = True

    def toggle_pause(self):
        self._paused = not self._paused
        self.events.sequencePauseToggled.emit(self._paused)

    def state(self, exclude=()) -> dict:
        """A dict with commonly accessed state values.  Faster than getSystemState."""
        # approx retrieval cost in comment (for demoCam)
        return {
            "AutoFocusDevice": self.getAutoFocusDevice(),  # 150 ns
            "BytesPerPixel": self.getBytesPerPixel(),  # 149 ns
            "CameraChannelNames": self.getCameraChannelNames(),  # 1 µs
            "CameraDevice": self.getCameraDevice(),  # 159 ns
            "Datetime": str(datetime.now()),
            "Exposure": self.getExposure(),  # 726 ns
            "FocusDevice": self.getFocusDevice(),  # 112 ns
            "GalvoDevice": self.getGalvoDevice(),  # 109 ns
            "ImageBitDepth": self.getImageBitDepth(),  # 147 ns
            "ImageHeight": self.getImageHeight(),  # 164 ns
            "ImageProcessorDevice": self.getImageProcessorDevice(),  # 110 ns
            "ImageWidth": self.getImageWidth(),  # 172 ns
            "PixelSizeUm": self.getPixelSizeUm(True),  # 2.2 µs  (True==cached)
            "ShutterDevice": self.getShutterDevice(),  # 152 ns
            "SLMDevice": self.getSLMDevice(),  # 110 ns
            "XYPosition": self.getXYPosition(),  # 1.1 µs
            "XYStageDevice": self.getXYStageDevice(),  # 156 ns
            "ZPosition": self.getZPosition(),  # 1.03 µs
        }


for name in (
    "getConfigData",
    "getConfigGroupState",
    "getConfigGroupStateFromCache",
    "getConfigState",
    "getSystemState",
    "getSystemStateCache",
):
    native_doc = getattr(pymmcore.CMMCore, name).__doc__
    getattr(CMMCorePlus, name).__doc__ += (
        "\n"
        + native_doc
        + dedent(
            """
    By default, this method returns a `pymmcore_plus.Configuration` object, which
    provides some conveniences over the native `pymmcore.Configuration` object, however
    this adds a little overhead. Use `native=True` to avoid the conversion.
    """
        ).strip()
    )


class _MMCallbackRelay(pymmcore.MMEventCallback):
    """Relays MMEventCallback methods to CMMCorePlus.signal."""

    def __init__(self, emitter: _CMMCoreSignaler):
        self._emitter = emitter
        super().__init__()

    @staticmethod
    def _make_reemitter(name):
        sig_name = name[2].lower() + name[3:]

        def reemit(self: _MMCallbackRelay, *args):
            try:
                getattr(self._emitter, sig_name).emit(*args)
            except Exception as e:
                import logging

                logging.getLogger(__name__).error(
                    "Exception occured in MMCorePlus callback %s: %s"
                    % (repr(sig_name), str(e))
                )

        return reemit


MMCallbackRelay = type(
    "MMCallbackRelay",
    (_MMCallbackRelay,),
    {
        n: _MMCallbackRelay._make_reemitter(n)
        for n in dir(pymmcore.MMEventCallback)
        if n.startswith("on")
    },
)



import atexit
import subprocess
import sys
import threading
import time
from typing import Optional

from loguru import logger
from Pyro5 import api, core, errors
from typing_extensions import Protocol

from .. import server
from .._serialize import register_serializers


class CallbackProtocol(Protocol):
    def receive_core_callback(self, signal_name: str, args: tuple) -> None:
        """Will be called by server with name of signal, and tuple of args."""


def _get_auto_callback_class():
    for modname in {"PyQt5", "PySide2", "PyQt6", "PySide6"}:
        qmodule = sys.modules.get(modname)
        if qmodule:
            QtWidgets = getattr(qmodule, "QtWidgets")
            if QtWidgets.QApplication.instance() is not None:
                from .callbacks.qcallback import QCoreCallback

                return QCoreCallback

    from .callbacks.basic import SynchronousCallback

    return SynchronousCallback


class RemoteMMCore(api.Proxy):
    def __init__(
        self,
        *,
        host: str = server.DEFAULT_HOST,
        port: int = server.DEFAULT_PORT,
        timeout: int = 5,
        verbose: bool = False,
        cleanup_new=True,
        connected_socket=None,
        callback_class=None,
    ):
        if callback_class is None:
            callback_class = _get_auto_callback_class()

        register_serializers()
        ensure_server_running(host, port, timeout, verbose, cleanup_new)

        uri = f"PYRO:{server.CORE_NAME}@{host}:{port}"
        super().__init__(uri, connected_socket=connected_socket)

        self.events = callback_class()
        cb_thread = DaemonThread(name="CallbackDaemon")
        cb_thread._daemon.register(self.events)
        self.connect_remote_callback(self.events)  # must come after register()
        cb_thread.start()

    def __getattr__(self, name):
        if name in ("events",):
            return object.__getattribute__(self, name)
        return super().__getattr__(name)

    def __setattr__(self, name, value):
        if name in ("events",):
            return object.__setattr__(self, name, value)
        return super().__setattr__(name, value)


def new_server_process(
    host: str, port: int, timeout=5, verbose=False
) -> subprocess.Popen:
    """Create a new daemon process"""
    cmd = [sys.executable, "-m", server.__name__, "-p", str(port), "--host", host]
    if verbose:
        cmd.append("--verbose")

    proc = subprocess.Popen(cmd)

    uri = f"PYRO:{core.DAEMON_NAME}@{host}:{port}"
    remote_daemon = api.Proxy(uri)

    while timeout > 0:
        try:
            remote_daemon.ping()
            return proc
        except Exception:
            timeout -= 0.1
            time.sleep(0.1)
    raise TimeoutError(f"Timeout connecting to server {uri}")


def ensure_server_running(
    host, port, timeout=5, verbose=False, cleanup_new=True
) -> Optional[subprocess.Popen]:
    """Ensure that a server daemon is running, or start one."""
    uri = f"PYRO:{core.DAEMON_NAME}@{host}:{port}"
    remote_daemon = api.Proxy(uri)
    try:
        remote_daemon.ping()
        logger.debug("Found existing server:\n{}", remote_daemon.info())
    except errors.CommunicationError:
        logger.debug("No server found, creating new mmcore server")
        proc = new_server_process(host, port, verbose=verbose)
        if cleanup_new:
            atexit.register(proc.kill)

            @atexit.register
            def _try_unload():
                try:
                    proxy = api.Proxy(f"PYRO:{server.CORE_NAME}@{host}:{port}")
                    proxy.unloadAllDevices()
                except errors.CommunicationError:
                    pass

        return proc
    return None


class DaemonThread(threading.Thread):
    def __init__(self, daemon=True, name="DaemonThread"):
        self._daemon = api.Daemon()
        self._stop_event = threading.Event()
        super().__init__(target=self._daemon.requestLoop, name=name, daemon=daemon)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.stop()



from Pyro5 import api

from ...core._signals import _CMMCoreSignaler


class SynchronousCallback(_CMMCoreSignaler):
    def __init__(self) -> None:
        super().__init__()

    @api.expose
    def receive_core_callback(self, signal_name: str, args: tuple) -> None:
        """Will be called by server with name of signal, and tuple of args."""
        getattr(self, signal_name).emit(*args)



import os
import re
import shutil
import ssl
import urllib.request
from pathlib import Path
from subprocess import run

_version_regex = re.compile(r"(\d+\.){2}\d+")
VERSION = "2.0.1"
RELEASE = 20211007
DEFAULT_DEST = Path(__file__).parent

ssl._create_default_https_context = ssl._create_unverified_context


def progressBar(current, chunksize, total, barLength=40):
    percent = float(current * chunksize) * 100 / total
    arrow = "-" * int(percent / 100 * barLength - 1) + ">"
    spaces = " " * (barLength - len(arrow))
    if not os.getenv("CI"):
        print("Progress: [%s%s] %d %%" % (arrow, spaces, percent), end="\r")


def download_url(url, output_path):
    print(f"downloading {url} ...")
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    urllib.request.urlretrieve(url, filename=output_path, reporthook=progressBar)


def _mac_main(dest_dir=DEFAULT_DEST, version=VERSION, release=RELEASE):
    if release == "latest":
        url = "https://download.micro-manager.org/latest/macos/"
        fname = "Micro-Manager-x86_64-latest.dmg"
        dst = dest_dir / "Micro-Manager-latest_mac"
    else:
        url = "https://valelab4.ucsf.edu/~MM/builds/2.0/Mac/"
        fname = f"Micro-Manager-{version}-{release}.dmg"
        dst = dest_dir / f"{fname[:-4]}_mac"

    if dst.exists():
        resp = input(f"Micro-manager already exists at\n{dst}\nOverwrite [Y/n]?")
        if resp.lower().startswith("n"):
            print("aborting")
            return

    download_url(f"{url}{fname}", fname)
    run(["hdiutil", "attach", "-nobrowse", fname], check=True)
    try:
        src = next(Path("/Volumes/Micro-Manager").glob("Micro-Manager*"))
    except StopIteration:
        src = f"/Volumes/Micro-Manager/{fname[:-4]}"
    shutil.copytree(src, dst, dirs_exist_ok=True)
    run(["hdiutil", "detach", "/Volumes/Micro-Manager"], check=True)
    os.unlink(fname)
    # fix gatekeeper ... requires password
    print(
        "\nYour password may be required to enable Micro-manager "
        "in your security settings."
        "\nNote: you can also quit now (cmd-C) and do this "
        "manually in the Security & Privacy preference pane."
    )
    run(["sudo", "xattr", "-r", "-d", "com.apple.quarantine", str(dst)], check=True)
    # # fix path randomization
    os.rename(dst / "ImageJ.app", "ImageJ.app")
    os.rename("ImageJ.app", dst / "ImageJ.app")
    return dst


def _win_main(dest_dir=DEFAULT_DEST, version=VERSION, release=RELEASE):
    if release == "latest":
        url = "https://download.micro-manager.org/latest/windows/"
        fname = "MMSetup_x64_latest.exe"
        dst = dest_dir / "Micro-Manager-latest_win"
    else:
        url = f"https://valelab4.ucsf.edu/~MM/nightlyBuilds/{version}/Windows/"
        dst = dest_dir / f"Micro-Manager-{version}-{release}_win"
        fname = f"MMSetup_64bit_{version}_{release}.exe"

    if dst.exists():
        resp = input(f"Micro-manager already exists at\n{dst}\nOverwrite [Y/n]?")
        if resp.lower().startswith("n"):
            print("aborting")
            return

    download_url(f"{url}{fname}", fname)
    run(
        [fname, "/VERYSILENT", "/SUPPRESSMSGBOXES", "/NORESTART", f"/DIR={dst}"],
        check=True,
    )
    os.unlink(fname)
    return dst


def install(dest_dir=DEFAULT_DEST, version=VERSION, release="latest"):
    out = (
        _win_main(dest_dir, version, release)
        if os.name == "nt"
        else _mac_main(dest_dir, version, release)
    )
    if out:
        print("installed to", out)


def _existing_dir(string):
    path = Path(string)
    if not path.is_dir():
        raise NotADirectoryError(string)
    return path


def _version(value: str):
    if not _version_regex.match(value):
        raise ValueError(
            f"Invalid version: {value}. Must be of form x.y.z with x y and z in 0-9"
        )
    return value


def _release(value: str):
    if value.lower() == "latest":
        return "latest"
    if len(value) != 8:
        raise ValueError(f"Invalid date: {value}. Must be eight digits.")
    return int(value)


def main():
    import sys

    print(sys.argv)
    import argparse

    parser = argparse.ArgumentParser(description="MM Device adapter installer.")
    parser.add_argument(
        "-d",
        "--dest",
        default=DEFAULT_DEST,
        type=_existing_dir,
        help=f"Directory in which to install (default: {DEFAULT_DEST})",
    )

    parser.add_argument(
        "-v",
        "--version",
        metavar="VERSION",
        type=_version,
        default=VERSION,
        help="Version number. e.g. 2.0.1 - ignored if release=latest "
        f"(default: {VERSION})",
    )
    parser.add_argument(
        "-r",
        "--release",
        metavar="DATE",
        type=_release,
        default="latest",
        help='8 digit date (YYYYMMDD) of MM nightly build to fetch, or "latest" '
        "(default: latest)",
    )

    args = parser.parse_args()
    install(args.dest, args.version, args.release)


if __name__ == "__main__":
    main()



import os
import re
import sys
from itertools import chain
from pathlib import Path
from typing import Optional, Type

camel_to_snake = re.compile(r"(?<!^)(?=[A-Z])")


def find_micromanager() -> Optional[str]:
    """Locate a Micro-Manager folder (for device adapters)."""
    from loguru import logger

    # environment variable takes precedence
    env_path = os.getenv("MICROMANAGER_PATH")
    if env_path and os.path.isdir(env_path):
        logger.debug(f"using MM path from env var: {env_path}")
        return env_path
    # then look for an installation in this folder (use `install_mm.sh` to install)
    sfx = "_win" if os.name == "nt" else "_mac"
    local_install = list(Path(__file__).parent.parent.glob(f"**/Micro-Manager*{sfx}"))
    if local_install:
        logger.debug(f"using MM path from local install: {local_install[0]}")
        return str(local_install[0])

    applications = {
        "darwin": Path("/Applications/"),
        "win32": Path("C:/Program Files/"),
        "linux": Path("/usr/local/lib"),
    }
    try:
        app_path = applications[sys.platform]
        pth = str(next(app_path.glob("[m,M]icro-[m,M]anager*")))
        logger.debug(f"using MM path found in applications: {pth}")
        return pth
    except KeyError:
        raise NotImplementedError(
            f"MM autodiscovery not implemented for platform: {sys.platform}"
        )
    except StopIteration:
        logger.error(f"could not find micromanager directory in {app_path}")
        return None


def wrap_for_pyro(cls: Type) -> Type:
    """Create proxy class compatible with pyro.

    Some classes, such as those autogenerated by SWIG, may be difficult
    to expose via `Pyro.api.expose`, because Pyro wants to add attributes
    directly to every class and class method that it exposes.  In some
    cases, this leads to an error like:

    AttributeError: 'method_descriptor' object has no attribute '_pyroExposed'

    This wrapper takes a class and returns a proxy class whose methods can
    all be modified.
    """

    def __init__(self, *args, **kwargs) -> None:
        self._obj = cls(*args, **kwargs)

    def _proxy_method(name):
        def _f(self, *args, **kwargs):
            obj = getattr(self, "_obj")
            method = getattr(obj, name)
            return method(*args, **kwargs)

        _f.__name__ = name
        return _f

    _dict_ = {}
    for k, v in chain(*(c.__dict__.items() for c in reversed(cls.mro()))):
        if callable(v) and not k.startswith("_"):
            _dict_[k] = _proxy_method(k)
            for attr in dir(v):
                if attr.startswith("_pyro"):
                    setattr(_dict_[k], attr, getattr(v, attr))

    _dict_["__init__"] = __init__
    return type(f"{cls.__name__}Proxy", (), _dict_)



import atexit
import datetime
from abc import ABC, abstractmethod
from multiprocessing.shared_memory import SharedMemory
from typing import Deque, Generic, TypeVar

import numpy as np
import pymmcore
import Pyro5
import Pyro5.api
import useq
from pydantic.datetime_parse import parse_duration

from .core import Configuration, Metadata

Pyro5.config.SERIALIZER = "msgpack"
T = TypeVar("T")


class Serializer(ABC, Generic[T]):
    # define these in subclasses

    @abstractmethod
    def to_dict(self, obj: T) -> dict:
        ...

    @abstractmethod
    def from_dict(self, classname: str, dct: dict) -> T:
        ...

    # -----------------

    @classmethod
    def type_(cls):
        return cls.__orig_bases__[0].__args__[0]

    def _to_dict(self, obj: T) -> dict:
        return {**self.to_dict(obj), "__class__": self.type_key()}

    def _from_dict(self, classname: str, d: dict) -> T:
        d.pop("__class__", None)
        return self.from_dict(classname, d)

    @classmethod
    def register(cls):
        ser = cls()
        Pyro5.api.register_class_to_dict(cls.type_(), ser._to_dict)
        Pyro5.api.register_dict_to_class(cls.type_key(), ser._from_dict)

    @classmethod
    def type_key(cls):
        return f"{cls.type_().__module__}.{cls.type_().__name__}"


class SerMDASequence(Serializer[useq.MDASequence]):
    def to_dict(self, obj: useq.MDASequence):
        return obj.dict()

    def from_dict(self, classname: str, d: dict):
        return useq.MDASequence.parse_obj(d)


class SerMDAEvent(Serializer[useq.MDAEvent]):
    def to_dict(self, obj: useq.MDAEvent):
        return obj.dict()

    def from_dict(self, classname: str, d: dict):
        return useq.MDAEvent.parse_obj(d)


class SerConfiguration(Serializer[Configuration]):
    def to_dict(self, obj: Configuration):
        return obj.dict()

    def from_dict(self, classname: str, d: dict):
        return Configuration.create(**d)


class SerMetadata(Serializer[Metadata]):
    def to_dict(self, obj: Metadata):
        return dict(obj)

    def from_dict(self, classname: str, d: dict):
        return Metadata(**d)


class SerTimeDelta(Serializer[datetime.timedelta]):
    def to_dict(self, obj: datetime.timedelta):
        return {"val": str(obj)}

    def from_dict(self, classname: str, d: dict):
        return parse_duration(d["val"])


class SerCMMError(Serializer[pymmcore.CMMError]):
    def to_dict(self, obj: pymmcore.CMMError):
        try:
            msg = obj.getMsg()
        except Exception:  # pragma: no cover
            msg = ""
        return {"msg": msg}

    def from_dict(self, classname: str, d: dict):
        return pymmcore.CMMError(str(d.get("msg")))


class SerNDArray(Serializer[np.ndarray]):
    SHM_SENT: Deque[SharedMemory] = Deque(maxlen=15)

    def to_dict(self, obj: np.ndarray):
        shm = SharedMemory(create=True, size=obj.nbytes)
        SerNDArray.SHM_SENT.append(shm)
        b: np.ndarray = np.ndarray(obj.shape, dtype=obj.dtype, buffer=shm.buf)
        b[:] = obj[:]
        return {
            "shm": shm.name,
            "shape": obj.shape,
            "dtype": str(obj.dtype),
        }

    def from_dict(self, classname: str, d: dict):
        """convert dict from `ndarray_to_dict` back to np.ndarray"""
        shm = SharedMemory(name=d["shm"], create=False)
        array = np.ndarray(d["shape"], dtype=d["dtype"], buffer=shm.buf).copy()
        shm.close()
        shm.unlink()
        return array


@atexit.register  # pragma: no cover
def _cleanup():
    for shm in SerNDArray.SHM_SENT:
        shm.close()
        try:
            shm.unlink()
        except FileNotFoundError:
            pass


def remove_shm_from_resource_tracker():
    """Monkey-patch multiprocessing.resource_tracker so SharedMemory won't be tracked

    More details at: https://bugs.python.org/issue38119
    """
    from multiprocessing import resource_tracker

    def fix_register(name, rtype):  # pragma: no cover
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(name, rtype)

    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):  # pragma: no cover
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(name, rtype)

    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


def register_serializers():
    remove_shm_from_resource_tracker()
    for i in globals().values():
        if isinstance(i, type) and issubclass(i, Serializer) and i != Serializer:
            i.register()



import json
import os
import re
from pathlib import Path
from threading import Thread
from unittest.mock import MagicMock, call, patch

import numpy as np
import psygnal
import pymmcore
import pytest
from pymmcore import CMMCore, PropertySetting
from qtpy.QtCore import Signal as QSignal
from useq import MDASequence

import pymmcore_plus
from pymmcore_plus import (
    CMMCorePlus,
    Configuration,
    DeviceDetectionStatus,
    DeviceType,
    Metadata,
    PropertyType,
)
from pymmcore_plus.core._signals import _CMMCoreSignaler


@pytest.fixture(params=["QSignal", "psygnal"], scope="function")
def core(request):
    core = CMMCorePlus()
    if request.param == "psygnal":
        core.events = _CMMCoreSignaler()
        core._callback_relay = pymmcore_plus.core._mmcore_plus.MMCallbackRelay(
            core.events
        )
        core.registerCallback(core._callback_relay)
    if not core.getDeviceAdapterSearchPaths():
        pytest.fail(
            "To run tests, please install MM with `python -m pymmcore_plus.install`"
        )
    core.loadSystemConfiguration()
    return core


def test_core(core: CMMCorePlus):
    assert isinstance(core, CMMCorePlus)
    assert isinstance(core, CMMCore)
    # because the fixture tries to find micromanager, this should be populated
    assert core.getDeviceAdapterSearchPaths()
    assert isinstance(core.events.propertyChanged, (psygnal.SignalInstance, QSignal))
    assert not core._canceled
    assert not core._paused

    # because the fixture loadsSystemConfig 'demo'
    assert len(core.getLoadedDevices()) == 12

    assert "CMMCorePlus" in repr(core)


def test_search_paths(core: CMMCorePlus):
    """Make sure search paths get added to path"""
    core.setDeviceAdapterSearchPaths(["test_path"])
    assert "test_path" in os.getenv("PATH")

    with pytest.raises(TypeError):
        core.setDeviceAdapterSearchPaths("test_path")


def test_load_system_config(core: CMMCorePlus):
    with pytest.raises(FileNotFoundError):
        core.loadSystemConfiguration("nonexistent")

    config_path = Path(__file__).parent / "local_config.cfg"
    core.loadSystemConfiguration(str(config_path))
    assert core.getLoadedDevices() == (
        "DHub",
        "Camera",
        "Dichroic",
        "Emission",
        "Excitation",
        "Objective",
        "Z",
        "Path",
        "XY",
        "Shutter",
        "Autofocus",
        "Core",
    )


def test_cb_exceptions(core: CMMCorePlus, caplog, qtbot):
    @core.events.propertyChanged.connect
    def _raze():
        raise ValueError("Boom")

    # using this to avoid our setProperty override... which would immediately
    # raise the exception (we want it to be raised deeper)
    if isinstance(core.events, _CMMCoreSignaler):
        pymmcore.CMMCore.setProperty(core, "Camera", "Binning", 2)
        msg = caplog.records[0].message
        assert msg == "Exception occured in MMCorePlus callback 'propertyChanged': Boom"
    else:
        with qtbot.capture_exceptions() as exceptions:
            with qtbot.waitSignal(core.events.propertyChanged):
                pymmcore.CMMCore.setProperty(core, "Camera", "Binning", 2)
        assert len(exceptions) == 1
        assert str(exceptions[0][1]) == "Boom"


def test_new_position_methods(core: CMMCorePlus):
    x1, y1 = core.getXYPosition()
    z1 = core.getZPosition()

    core.setRelativeXYZPosition(1, 1, 1)

    x2, y2 = core.getXYPosition()
    z2 = core.getZPosition()

    assert round(x2, 2) == x1 + 1
    assert round(y2, 2) == y1 + 1
    assert round(z2, 2) == z1 + 1


def test_mda(core: CMMCorePlus, qtbot):
    """Test signal emission during MDA"""
    mda = MDASequence(
        time_plan={"interval": 0.1, "loops": 2},
        stage_positions=[(1, 1, 1)],
        z_plan={"range": 3, "step": 1},
        channels=[{"config": "DAPI", "exposure": 1}],
    )
    fr_mock = MagicMock()
    ss_mock = MagicMock()
    sf_mock = MagicMock()
    xystage_mock = MagicMock()
    stage_mock = MagicMock()
    exp_mock = MagicMock()

    core.events.frameReady.connect(fr_mock)
    core.events.sequenceStarted.connect(ss_mock)
    core.events.sequenceFinished.connect(sf_mock)
    core.events.XYStagePositionChanged.connect(xystage_mock)
    core.events.stagePositionChanged.connect(stage_mock)
    core.events.exposureChanged.connect(exp_mock)

    with qtbot.waitSignal(core.events.sequenceFinished):
        core.run_mda(mda)
    assert fr_mock.call_count == len(list(mda))
    for event, _call in zip(mda, fr_mock.call_args_list):
        assert isinstance(_call.args[0], np.ndarray)
        assert _call.args[1] == event

    ss_mock.assert_called_once_with(mda)
    sf_mock.assert_called_once_with(mda)
    xystage_mock.assert_called_with("XY", 1.0, 1.0)
    exp_mock.assert_called_with("Camera", 1.0)
    stage_mock.assert_has_calls(
        [
            call("Z", -0.5),
            call("Z", 0.5),
            call("Z", 1.5),
            call("Z", 2.5),
            call("Z", -0.5),
            call("Z", 0.5),
            call("Z", 1.5),
            call("Z", 2.5),
        ]
    )


def test_mda_pause_cancel(core: CMMCorePlus, qtbot):
    """Test signal emission during MDA with cancelation"""
    mda = MDASequence(
        time_plan={"interval": 0.1, "loops": 2},
        stage_positions=[(1, 1, 1)],
        z_plan={"range": 3, "step": 1},
        channels=[{"config": "DAPI", "exposure": 1}],
    )

    pause_mock = MagicMock()
    cancel_mock = MagicMock()
    sf_mock = MagicMock()
    ss_mock = MagicMock()

    core.events.sequenceStarted.connect(ss_mock)
    core.events.sequencePauseToggled.connect(pause_mock)
    core.events.sequenceCanceled.connect(cancel_mock)
    core.events.sequenceFinished.connect(sf_mock)

    _fcount = 0

    @core.events.frameReady.connect
    def _onframe(frame, event):
        nonlocal _fcount
        _fcount += 1
        if _fcount == 2:
            core.toggle_pause()
            pause_mock.assert_called_with(True)
            core.toggle_pause()
            pause_mock.assert_called_with(False)
        elif _fcount == 4:
            core.cancel()

    with qtbot.waitSignal(core.events.sequenceFinished):
        core.run_mda(mda)

    ss_mock.assert_called_once_with(mda)
    cancel_mock.assert_called_once_with(mda)
    assert _fcount == 4
    assert _fcount < len(list(mda))
    sf_mock.assert_called_once_with(mda)


def test_device_type_overrides(core: CMMCorePlus):
    dt = core.getDeviceType("Camera")
    assert isinstance(dt, DeviceType)
    assert str(dt) == "Camera"
    assert int(dt) == 2
    assert dt == DeviceType["Camera"]
    assert dt == DeviceType["CameraDevice"]
    assert dt == DeviceType(2)


def test_property_type_overrides(core: CMMCorePlus):
    pt = core.getPropertyType("Camera", "Binning")
    assert isinstance(pt, PropertyType)
    assert pt.to_python() is int


def test_detect_device(core: CMMCorePlus):
    dds = core.detectDevice("Camera")
    assert isinstance(dds, DeviceDetectionStatus)
    assert dds == -2 == DeviceDetectionStatus.Unimplemented


def test_metadata(core: CMMCorePlus):
    core.startContinuousSequenceAcquisition(10)
    core.stopSequenceAcquisition()
    image, md = core.getLastImageMD()
    assert isinstance(md, Metadata)
    assert md["Height"] == "512"
    assert "Binning" in md.keys()
    assert ("ImageNumber", "0") in md.items()
    assert "GRAY16" in md.values()

    assert "Camera" in md
    md["Camera"] = "new"
    assert md["Camera"] == "new" == md.get("Camera")

    cpy = md.copy()
    assert cpy == md

    del md["Camera"]
    assert "Camera" not in md

    assert "Camera" in cpy
    assert md.get("", 1) == 1  # default

    md.clear()
    assert not md

    assert isinstance(md.json(), str)


def test_new_metadata():
    md = Metadata({"a": "1"})
    assert md["a"] == "1"
    assert isinstance(md, pymmcore.Metadata)


def test_md_overrides(core: CMMCorePlus):
    core.startContinuousSequenceAcquisition(10)
    core.stopSequenceAcquisition()

    image, md = core.getNBeforeLastImageMD(0)
    assert isinstance(md, Metadata)

    image, md = core.popNextImageMD()
    assert isinstance(md, Metadata)


def test_configuration(core: CMMCorePlus):
    state = core.getSystemState()
    assert isinstance(state, Configuration)
    assert not isinstance(core.getSystemState(native=True), Configuration)

    assert str(state)

    tup = tuple(state)
    assert isinstance(tup, tuple)
    assert all(isinstance(x, tuple) and len(x) == 3 for x in tup)

    with pytest.raises(TypeError):
        assert state["Camera"] == 1
    with pytest.raises(TypeError):
        assert "Camera" in state

    assert state["Camera", "Binning"] == "1"
    assert PropertySetting("Camera", "Binning", "1") in state
    assert state in state

    assert ("Camera", "Binning") in state


def test_config_create():
    input = {"a": {"a0": "0", "a1": "1"}, "b": {"b0": "10", "b1": "11"}}
    aslist = [(d, p, v) for d, ps in input.items() for p, v in ps.items()]
    cfg1 = Configuration.create(input)
    cfg2 = Configuration.create(aslist)
    cfg3 = Configuration.create(a=input["a"], b=input["b"])
    assert cfg1.dict() == cfg2.dict() == cfg3.dict() == input
    assert list(cfg1) == list(cfg2) == list(cfg3) == aslist
    assert cfg1 == cfg2 == cfg3

    assert cfg1.json() == json.dumps(input)
    assert cfg1.html()


def test_config_yaml():
    input = {"a": {"a0": "0", "a1": "1"}, "b": {"b0": "10", "b1": "11"}}
    cfg1 = Configuration.create(input)
    yaml = pytest.importorskip("yaml")
    assert cfg1.yaml() == yaml.safe_dump(input)


def test_property_schema(core: CMMCorePlus):
    schema = core.getDeviceSchema("Camera")
    assert isinstance(schema, dict)
    assert schema["title"] == "DCam"
    assert schema["properties"]["AllowMultiROI"] == {"type": "boolean"}


def test_set_property_events(core: CMMCorePlus):
    """Test that using setProperty always emits a propertyChanged event."""
    mock = MagicMock()
    core.events.propertyChanged.connect(mock)
    core.setProperty("Camera", "Binning", "2")
    mock.assert_called_once_with("Camera", "Binning", "2")

    mock.reset_mock()
    core.setProperty("Camera", "Binning", "1")
    mock.assert_called_once_with("Camera", "Binning", "1")

    mock.reset_mock()
    core.setProperty("Camera", "Binning", "1")
    mock.assert_not_called()  # value didn't change

    # this is not a property that the DemoCamera emits...
    # so with regular pymmcore, this would not be emitted.
    core.setProperty("Camera", "AllowMultiROI", "1")
    mock.assert_called_once_with("Camera", "AllowMultiROI", "1")


def test_get_objectives(core: CMMCorePlus):
    devices = core.guessObjectiveDevices()
    assert len(devices) == 1
    assert devices[0] == "Objective"

    with pytest.raises(TypeError):
        core.objective_device_pattern = 4

    # assign a new regex that won't match Objective using a str
    core.objective_device_pattern = "^((?!Objective).)*$"
    assert "Objective" not in core.guessObjectiveDevices()

    # assign new using a pre-compile pattern
    core.objective_device_pattern = re.compile("Objective")
    devices = core.guessObjectiveDevices()
    assert len(devices) == 1
    assert devices[0] == "Objective"


def test_guess_channel_group(core: CMMCorePlus):

    chan_group = core.getChannelGroup()
    assert chan_group == "Channel"

    assert core.getOrGuessChannelGroup() == ["Channel"]

    with patch.object(core, "getChannelGroup", return_value=""):
        assert core.getOrGuessChannelGroup() == ["Channel"]

        with pytest.raises(TypeError):
            core.channelGroup_pattern = 4

        # assign a new regex that won't match Channel using a str
        # this will return all the mm groups, but that's because this a bad regex
        # to use
        core.channelGroup_pattern = "^((?!(Channel)).)*$"
        assert core.getOrGuessChannelGroup() == [
            "Camera",
            "LightPath",
            "Objective",
            "System",
        ]

        # assign new using a pre-compile pattern
        core.channelGroup_pattern = re.compile("Channel")
        chan_group = core.getOrGuessChannelGroup()
        assert chan_group == ["Channel"]


def test_lock_and_callbacks(core: CMMCorePlus, qtbot):
    # when a function with a lock triggers a callback
    # that callback should be able to call locked functions
    # without hanging.

    # do some threading silliness here so we don't accidentally hang our
    # test if things go wrong have to use *got_lock* to check because we
    # can't assert in the function as theads don't throw their exceptions
    # back into the calling thread.
    got_lock = False

    def cb(*args, **kwargs):
        nonlocal got_lock
        got_lock = core.lock.acquire(timeout=0.1)
        if got_lock:
            core.lock.release()

    core.events.XYStagePositionChanged.connect(cb)

    def trigger_cb():
        core.setXYPosition(4, 5)

    th = Thread(target=trigger_cb)
    with qtbot.waitSignal(core.events.XYStagePositionChanged):
        th.start()
    assert got_lock
    got_lock = False

    core.events.frameReady.connect(cb)
    mda = MDASequence(
        time_plan={"interval": 0.1, "loops": 2},
        stage_positions=[(1, 1, 1)],
        z_plan={"range": 3, "step": 1},
        channels=[{"config": "DAPI", "exposure": 1}],
    )

    with qtbot.waitSignal(core.events.sequenceFinished):
        core.run_mda(mda)
    assert got_lock


def test_single_instance():
    core1 = CMMCorePlus.instance()
    core2 = CMMCorePlus.instance()
    assert core1 is core2



from enum import IntEnum, auto

# Enums


class DeviceType(IntEnum):
    UnknownType = 0
    AnyType = auto()
    CameraDevice = auto()
    ShutterDevice = auto()
    StateDevice = auto()
    StageDevice = auto()
    XYStageDevice = auto()
    SerialDevice = auto()
    GenericDevice = auto()
    AutoFocusDevice = auto()
    CoreDevice = auto()
    ImageProcessorDevice = auto()
    SignalIODevice = auto()
    MagnifierDevice = auto()
    SLMDevice = auto()
    HubDevice = auto()
    GalvoDevice = auto()
    # aliases for clearer naming (e.g. `DeviceType.Camera`)
    Unknown = UnknownType
    Any = AnyType
    Camera = CameraDevice
    Shutter = ShutterDevice
    State = StateDevice
    Stage = StageDevice
    XYStage = XYStageDevice
    Serial = SerialDevice
    Generic = GenericDevice
    AutoFocus = AutoFocusDevice
    Core = CoreDevice
    ImageProcessor = ImageProcessorDevice
    SignalIO = SignalIODevice
    Magnifier = MagnifierDevice
    SLM = SLMDevice
    Hub = HubDevice
    Galvo = GalvoDevice

    def __str__(self):
        return self.name.replace("Type", "").replace("Device", "")


class PropertyType(IntEnum):
    Undef = 0
    String = auto()
    Float = auto()
    Integer = auto()

    def to_python(self):
        return {0: None, 1: str, 2: float, 3: int}[self]

    def to_json(self):
        return {0: "null", 1: "string", 2: "number", 3: "integer"}[self]


class ActionType(IntEnum):
    NoAction = 0
    BeforeGet = auto()
    AfterSet = auto()
    IsSequenceable = auto()
    AfterLoadSequence = auto()
    StartSequence = auto()
    StopSequence = auto()


class PortType(IntEnum):
    InvalidPort = 0
    SerialPort = auto()
    USBPort = auto()
    HIDPort = auto()


class FocusDirection(IntEnum):
    FocusDirectionUnknown = 0
    FocusDirectionTowardSample = auto()
    FocusDirectionAwayFromSample = auto()
    # aliases
    Unknown = FocusDirectionUnknown
    TowardSample = FocusDirectionTowardSample
    AwayFromSample = FocusDirectionAwayFromSample


class DeviceNotification(IntEnum):
    Attention = 0
    Done = auto()
    StatusChanged = auto()


class DeviceDetectionStatus(IntEnum):
    """DeviceDetectionStatus from device discovery

    Unimplemented
        there is as yet no mechanism to programmatically detect the device
    Misconfigured
        some information needed to communicate with the device is invalid
    CanNotCommunicate
        communication attributes are valid, but the device does not respond
    CanCommunicate
        communication verified, parameters have been set to valid values.
    """

    Unimplemented = -2
    Misconfigured = -1
    CanNotCommunicate = 0
    CanCommunicate = 1



from functools import partial
from typing import TYPE_CHECKING, Set

from Pyro5 import errors
from Pyro5.api import behavior, expose, oneway

from .._util import wrap_for_pyro
from ..core._mmcore_plus import CMMCorePlus
from ..core._signals import _CMMCoreSignaler

if TYPE_CHECKING:
    from ..client._client import CallbackProtocol


_SIGNAL_NAMES = {name for name in dir(_CMMCoreSignaler) if not name.startswith("_")}


@expose
@behavior(instance_mode="single")
@wrap_for_pyro
class pyroCMMCore(CMMCorePlus):
    """CMMCorePlus instance running on a server.

    It should mostly mimic CMMCorePlus but it may have some additional characteristics
    that are only necessary for asynchronous/remote usage (like callbacks)
    """

    _callback_handlers: Set["CallbackProtocol"] = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name in _SIGNAL_NAMES:
            getattr(self.events, name).connect(partial(self.emit_signal, name))

    def connect_remote_callback(self, handler: "CallbackProtocol"):
        self._callback_handlers.add(handler)

    def disconnect_remote_callback(self, handler: "CallbackProtocol"):
        self._callback_handlers.discard(handler)

    @oneway
    def run_mda(self, sequence) -> None:  # pragma: no cover
        return super().run_mda(sequence)

    @oneway
    def emit_signal(self, signal_name: str, *args):
        from loguru import logger

        logger.debug("{}: {}", signal_name, args)
        for handler in list(self._callback_handlers):
            try:
                handler._pyroClaimOwnership()  # type: ignore
                handler.receive_core_callback(signal_name, args)
            except errors.CommunicationError:  # pragma: no cover
                self.disconnect_remote_callback(handler)



"""
An example of integrating a remote MMCore object into an event loop.

This example requires qtpy and an Qt backend installed in the env:
```
pip install qtpy pyqt5
```
"""
import numpy as np
from qtpy.QtWidgets import QApplication, QPushButton
from useq import MDAEvent, MDASequence

from pymmcore_plus import RemoteMMCore

app = QApplication([])

# button for early stopping
stop = QPushButton("STOP")
stop.clicked.connect(app.quit)
stop.show()

# see https://github.com/tlambert03/useq-schema
sequence = MDASequence(
    channels=["DAPI", {"config": "FITC", "exposure": 50}],
    time_plan={"interval": 1.5, "loops": 5},
    z_plan={"range": 4, "step": 0.5},
    axis_order="tpcz",
)

# start server in another process and connect to it
with RemoteMMCore() as mmcore:

    @mmcore.frameReady.connect
    def on_frame(image: np.ndarray, event: MDAEvent):
        print(
            f"received frame: {image.shape}, {image.dtype} "
            f"@ index {event.index}, z={event.z_pos}"
        )

    @mmcore.propertyChanged.connect
    def prop_changed(device, prop, value):
        print(f"{device}.{prop} changed to {value!r}")

    # setup some callbacks
    mmcore.systemConfigurationLoaded.connect(lambda: print("config loaded!"))
    mmcore.sequenceFinished.connect(app.quit)

    # load config and start an experiment
    mmcore.loadSystemConfiguration("demo")
    mmcore.run_mda(sequence)

    # start the qt event loop
    app.exec_()



import socket
import threading
from subprocess import Popen
from typing import Any, Type

from Pyro5 import api
from typing_extensions import Protocol as Protocol

from .._serialize import register_serializers as register_serializers
from ..core import CMMCorePlus

class CallbackProtocol(Protocol):
    def receive_core_callback(self, signal_name: str, args: tuple) -> None: ...

# doesn't actually subclass, but has same methods
class RemoteMMCore(api.Proxy, CMMCorePlus):
    def __init__(
        self,
        *,
        host=...,
        port=...,
        timeout: int = ...,
        verbose: bool = ...,
        cleanup_new: bool = ...,
        cleanup_existing: bool = ...,
        connected_socket: socket.socket | None = ...
    ) -> None: ...
    def _register_callback(self, callback: CallbackProtocol): ...

def new_server_process(
    host: str, port: int, timeout: int = ..., verbose: bool = ...
) -> Popen: ...
def ensure_server_running(
    host: str,
    port: int,
    timeout: int = ...,
    verbose: bool = ...,
    cleanup_new: bool = ...,
    cleanup_existing: bool = ...,
) -> Popen | None: ...

class DaemonThread(threading.Thread):
    def __init__(self, daemon: bool = ...) -> None: ...
    def __enter__(self): ...
    def __exit__(self, *args, **kwargs) -> None: ...
    _daemon: api.Daemon
    _stop_event: threading.Event



from useq import MDASequence

from pymmcore_plus import CMMCorePlus

# see https://github.com/tlambert03/useq-schema
sequence = MDASequence(
    channels=["DAPI", {"config": "FITC", "exposure": 50}],
    time_plan={"interval": 2, "loops": 5},
    z_plan={"range": 4, "step": 0.5},
    axis_order="tpcz",
)

mmc = CMMCorePlus()
mmc.loadSystemConfiguration()
mda_thread = mmc.run_mda(sequence)



from unittest.mock import MagicMock, patch

from pymmcore_plus.server import pyroCMMCore, serve


def test_server():
    core = pyroCMMCore()
    core.loadSystemConfiguration()

    assert core.getDeviceAdapterSearchPaths()
    cb = MagicMock()
    core.connect_remote_callback(cb)

    core.emit_signal("propertiesChanged")
    core.disconnect_remote_callback(cb)


def test_serve(monkeypatch):
    import sys

    monkeypatch.setattr(sys, "argv", ["serve", "-p", "65111"])
    with patch("Pyro5.api.serve") as mock:
        serve()
    mock.assert_called_once()



__all__ = [
    "CMMCorePlus",
    "Configuration",
    "Metadata",
    "DeviceType",
    "PropertyType",
    "ActionType",
    "PortType",
    "FocusDirection",
    "DeviceNotification",
    "DeviceDetectionStatus",
]

from ._config import Configuration
from ._constants import (
    ActionType,
    DeviceDetectionStatus,
    DeviceNotification,
    DeviceType,
    FocusDirection,
    PortType,
    PropertyType,
)
from ._metadata import Metadata
from ._mmcore_plus import CMMCorePlus



"""pythonic wrapper on pymmcore.Metadata object."""
from collections.abc import Mapping
from types import new_class
from typing import Any, ItemsView, Iterator, KeysView, ValuesView

import pymmcore

_NULL = object()


class Metadata(pymmcore.Metadata):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        if args and isinstance(args[0], Mapping):
            for k, v in args[0].items():
                self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def __getitem__(self, name: str) -> Any:
        try:
            return self.GetSingleTag(name).GetValue()
        except ValueError:
            raise KeyError(str(name))

    def __setitem__(self, name: str, value: Any) -> None:
        tag = pymmcore.MetadataSingleTag(name, "_", False)
        tag.SetValue(str(value))
        self.SetTag(tag)

    def __delitem__(self, name: str) -> None:
        self.RemoveTag(name)

    def __iter__(self) -> Iterator[str]:
        yield from self.GetKeys()

    def __contains__(self, tag: str):
        return self.HasTag(tag)

    def __len__(self) -> int:
        return len(self.GetKeys())

    def __repr__(self):
        return f"{dict(self)!r}"

    def get(self, name, default=_NULL):
        try:
            return self.__getitem__(name)
        except KeyError:
            if default is not _NULL:
                return default
            raise

    def copy(self):
        return type(self)(**dict(self))

    def clear(self) -> None:
        self.Clear()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Metadata):
            return False
        return dict(self) == dict(other)

    def json(self) -> str:
        import json

        return json.dumps(dict(self))

    def keys(self) -> KeysView[str]:
        return metadata_keys(self)

    def items(self) -> ItemsView[str, str]:
        return metadata_items(self)

    def values(self) -> ValuesView[str]:
        return metadata_values(self)


metadata_keys = new_class("metadata_keys", (KeysView,), {})
metadata_items = new_class("metadata_items", (ItemsView,), {})
metadata_values = new_class("metadata_values", (ValuesView,), {})



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from psutil import Process
CORE_NAME = "pymmcore_plus.CMMCorePlus"
DEFAULT_PORT = 54333
DEFAULT_HOST = "127.0.0.1"
DEFAULT_URI = f"PYRO:{CORE_NAME}@{DEFAULT_HOST}:{DEFAULT_PORT}"
VERBOSE = False


def _get_remote_pid(host, port) -> Optional["Process"]:
    import psutil

    for proc in psutil.process_iter(["connections"]):
        for pconn in proc.info["connections"] or []:
            if pconn.laddr.port == port and pconn.laddr.ip == host:
                return proc
    return None


def try_kill_server(host: str = None, port: int = None):
    from loguru import logger

    if host is None:
        host = DEFAULT_HOST
    if port is None:
        port = DEFAULT_PORT

    proc = _get_remote_pid(host, port)
    if proc is not None:
        proc.kill()
        logger.info(f"Killed process on {host=}:{port=}")
    else:
        logger.info("No process found")


def serve():
    import argparse

    import Pyro5
    from loguru import logger

    from pymmcore_plus._serialize import register_serializers
    from pymmcore_plus.server import pyroCMMCore

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", type=int, default=DEFAULT_PORT, help="port")
    parser.add_argument("--host", default=DEFAULT_HOST)
    parser.add_argument("--verbose", action="store_true", default=VERBOSE)
    args = parser.parse_args()

    if not args.verbose:
        logger.disable("pymmcore_plus")

    register_serializers()
    Pyro5.api.serve(
        {pyroCMMCore: CORE_NAME},
        use_ns=False,
        host=args.host,
        port=args.port,
        verbose=args.verbose,
    )


if __name__ == "__main__":
    serve()



from pymmcore_plus import RemoteMMCore

with RemoteMMCore(verbose=True) as mmcore:
    # 'demo' is a special option for the included CMMCorePlus
    # that loads the micro-manager demo config
    mmcore.loadSystemConfiguration("demo")
    print("loaded:", mmcore.getLoadedDevices())
